import os
from typing import Optional, Union
from multidict import CIMultiDict

from pyxk import (
    md5,
    rename,
    chardet,
    Session,
    normpath,
    make_open,
    default_headers,
    read_file_by_generator
)


class _M3U8Init:

    def __init__(
        self,
        store: Optional[str] = None,
        verify: bool = True,
        headers: Union[dict, CIMultiDict] = default_headers(),
        user_agent: Optional[str] = None,
        reload: bool = False,
        reserve: bool = False,
        limit: int = 16,
    ):
        """m3u8解析器

        :param store: m3u8文件 保存文件夹路径
        :param verify: verify ssl
        :param headers: Headers
        :param user_agent: User agent
        :param reload: 重载m3u8文件
        :param reserve: 下载完成保留m3u8文件
        :param limit: 异步下载limit
        """
        # 存储文件夹
        self._store = self._norm_store(store)
        # verify ssl
        self._verify = bool(verify)
        # headers
        self._headers = CIMultiDict(headers)
        # user agent
        if user_agent and isinstance(user_agent, str):
            self._headers['User-Agent'] = user_agent
        else:
            user_agent = None
        # reload 重载m3u8文件
        self._reload = bool(reload)
        # 保留m3u8文件
        self._reserve = bool(reserve)
        # 异步下载limit
        self._limit = limit
        # request session
        self._session = Session(headers=headers, user_agent=user_agent)
        # 文件名称
        self._output = None
        # 网络访问历史记录
        self.history = []

    @property
    def store(self) -> Optional[str]:
        """文件保存文件夹"""
        return self._store

    @store.setter
    def store(self, value):
        self._store = self._norm_store(value)

    @staticmethod
    def _norm_store(value) -> Optional[str]:
        """输入文件夹路径"""
        if value is None:
            return normpath(os.getcwd())

        # store 类型无效
        if not isinstance(value, str) or not value:
            raise TypeError(f'store must be a str, got {value!r}')

        return normpath(os.path.abspath(normpath(value)))

    @property
    def output(self) -> Optional[str]:
        """m3u8文件保存路径"""
        return self._output

    @output.setter
    def output(self, value):
        self._output = self._norm_output(value)

    def _norm_output(self, value, suffix='mp4') -> Optional[str]:
        """输入文件"""
        # 没有output
        if value is None or not value:
            return None

        # output 类型无效
        if not isinstance(value, str):
            raise TypeError(f'output must be a str, got {value!r}')

        # 绝对路径
        value = normpath(value)
        if not os.path.isabs(value):
            value = os.path.join(self.store, value)

        rename_file = rename(value, suffix=suffix)
        self._store = rename_file.dirname
        return rename_file.result

    @property
    def temp(self) -> Optional[str]:
        """m3u8 临时文件夹"""
        if not self.output:
            return None

        dir_name, basename = os.path.split(self.output)
        return normpath(os.path.join(dir_name, basename.removesuffix('mp4') + 'temp'))

    @property
    def _temp_segments(self) -> Optional[str]:
        """segments保存临时文件夹"""
        if not self.temp:
            return None

        return normpath(os.path.join(self.temp, 'segments'))

    @property
    def user_agent(self) -> str:
        """user agent string"""
        return self._headers.get('User-Agent')

    @user_agent.setter
    def user_agent(self, value):
        if not isinstance(value, str):
            raise ValueError(f'user_agent must be a str, got {value!r}')

        self._headers.update({'User-Agent': value})
        self.session.user_agent = value

    @property
    def headers(self) -> Union[dict, CIMultiDict]:
        """Request Headers"""
        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers.update(dict(value))
        self._session.headers.update(self._headers)

    @property
    def limit(self) -> int:
        """download limit"""
        return self._limit

    @limit.setter
    def limit(self, value):
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f'limit must be a positive integer(正整数), got {value!r}')

        self._limit = value

    @property
    def session(self):
        """requests session"""
        return self._session

    @property
    def console(self):
        """rich console"""
        return self._session.console

    def _initialization(
        self, verify, headers, user_agent, reload, reserve, limit
    ):
        """初始化变量"""
        if verify is not None:
            self._verify = bool(verify)
        if headers is not None:
            self.headers = headers
        if user_agent is not None:
            self.user_agent = user_agent
        if reload is not None:
            self._reload = bool(reload)
        if reserve is not None:
            self._reserve = bool(reversed)
        if limit is not None:
            self.limit = limit

    def generate_filename(self, url: Optional[str], is_key: bool = False) -> Optional[str]:
        """生成文件名称

        :param url: m3u8 url
        :param is_key: 是否为m3u8密钥
        :return:
        """
        if not self.output or not url:
            return None

        # 加密后的文件名称
        basename = md5(url).ciphertext + ('.key' if is_key else '.m3u8')
        return normpath(os.path.join(self.temp, basename))

    def get_m3u8_content(self, url: Optional[str], is_key: bool = False):
        """获取 m3u8 内容

        :param url: m3u8 链接
        :param is_key: 是否为m3u8密钥
        :return:
        """
        # 获取文件名称
        filename = self.generate_filename(url, is_key)

        if self._reload or not filename or not os.path.isfile(filename):
            return self._content_from_url(url, is_key)

        return self._content_from_file(filename, is_key)

    def _content_from_url(self, url: str, is_key: bool = False) -> Union[str, bytes]:
        """获取网络资源 - m3u8 content

        :param url: m3u8 url
        :param is_key: 是否为m3u8密钥
        :return: Union[str, bytes]
        """

        content = b''
        response = self.session.get(url, verify=self._verify, timeout=10, stream=True)
        self.history.append(response)

        for index, chunk in enumerate(response.iter_content(1024)):
            # 检查是否是m3u8内容
            if not is_key and index == 0 and not self._is_m3u8(chunk):
                break
            content += chunk

        # m3u8 content 编码转换
        if not is_key:
            content = content.decode(
                chardet(content).encoding or 'utf-8'
            )

        return content

    def _content_from_file(self, file: str, is_key: bool = False) -> Union[str, bytes]:
        """获取本地资源 - m3u8 content

        :param file: m3u8 file
        :param is_key: 是否为m3u8密钥
        :return: Union[str, bytes]
        """
        # 获取本地文件
        content = b'' if is_key else ''

        for index, chunk in enumerate(
            read_file_by_generator(
                file=file,
                mode='rb' if is_key else 'r',
                encoding=None if is_key else 'utf-8'
            )
        ):
            # 检查是否是m3u8内容
            if not is_key and index == 0 and not self._is_m3u8(chunk):
                break
            content += chunk

        return content

    def sava_m3u8_content(self, url: Optional[str], content: Union[str, bytes], is_key: bool = False):
        """保存m3u8文件到本地

        :param url: m3u8 url
        :param content: m3u8 content
        :param is_key: 是否为m3u8密钥
        :return:
        """
        filename = self.generate_filename(url, is_key)

        # 无效文件
        if not filename:
            return

        # 保存文件
        if self._reload or not os.path.isfile(filename):
            with make_open(
                file=filename,
                mode='wb' if is_key else 'w',
                encoding=None if is_key else 'utf-8'
            ) as write_file_obj:
                write_file_obj.write(content)

    @staticmethod
    def _is_m3u8(_data: Union[str, bytes], /) -> bool:
        """检查是否是m3u8文本

        :param _data: content
        :return: bool
        """
        if isinstance(_data, str):
            return _data.startswith('#EXTM3U')

        if isinstance(_data, bytes):
            return _data.startswith(b'#EXTM3U')

        return False



import os
import shlex
import subprocess
from typing import Optional
from concurrent.futures import ThreadPoolExecutor

import aiofiles

from pyxk import (
    Client,
    Response,
    tasks_progress,
    download_progress
)
from pyxk.aes import Crypto, MODES


class Downloader(Client):

    limit = 1000
    timeout = 100
    warning = False
    status_error_list = list(range(400, 411))
    until_request_succeed = [200]

    def __init__(
        self,
        table: object,
        output: str,
        m3u8keys: dict,
        start_urls: list,
        console: object,
        temp: str,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.ciphers = {
            url: Crypto(keys['key'], MODES.CBC, iv=keys['iv'])
            for url, keys in m3u8keys.items()
        }
        self.table = table
        self.output = output
        self.start_urls = start_urls

        # 进度条
        self.progress = tasks_progress()
        self.task = self.progress.add_task('', total=len(self.start_urls))

        # 添加进度条到live
        self.table.add_section()
        self.table.add_row(self.progress)

        self.temp = temp
        self.console = console

    async def start_request(self):

        ret = []
        for item in self.start_urls.copy():
            file = item[1]['file']
            ret.append(file)

            # 本地存在文件
            if os.path.isfile(file) and os.path.getsize(file) > 0:
                self.start_urls.remove(item)
                continue

        # 更新进度条
        self.progress.update(
            self.task,
            completed=len(ret)-len(self.start_urls)
        )

        # 没有下载数据
        if not self.start_urls:
            return ret

        await super().start_request()
        return ret

    async def parse(self, response: Response, file: str, key: Optional[str]):
        """解析segment数据"""
        # 获取segment内容
        content = await response.content.read()
        # 解密
        if key:
            content = self.ciphers[key].decrypt(content)

        async with aiofiles.open(file, 'wb') as write_file_obj:
            # 本地存储
            await write_file_obj.write(content)

        # 更新进度条
        self.progress.update(self.task, advance=1)

    async def completed(self, ret: list):
        """使用模块ffmpeg合并m3u8"""
        if not ret:
            return

        # 创建 filelist 文件
        filelist, filesize = os.path.join(self.temp, "filelist.txt"), 0

        with open(filelist, "w", encoding="utf-8") as write_fileobj:
            for file in ret:
                write_fileobj.write(f"file '{file}'\n")
                filesize += (os.path.getsize(file) - 16400)

        # ffmpeg 视频合并代码, 监测合并完成状态
        args = shlex.split(
            f"ffmpeg -loglevel quiet -f concat -safe 0 -i {filelist} -c copy {self.output} -y"
        )
        merge_completed = False

        # ffmpeg 合并函数
        def merge_segments():
            try:
                subprocess.run(args=args, check=True)
            except FileNotFoundError:
                self.console.log("[red]ffmpeg is not available![/]")
            finally:
                nonlocal merge_completed
                merge_completed = True

        # 合并进度条
        def merge_progress():
            from time import sleep

            sleep(0.2)
            if merge_completed:
                return

            completed_filezise = lambda :os.path.getsize(self.output) if os.path.isfile(self.output) else 0
            progress = download_progress(transient=True)
            task = progress.add_task(
                description="", total=filesize, visible=False
            )
            self.table.add_row(progress)

            while True:
                # 停止循环
                if merge_completed:
                    progress.update(task, completed=filesize, visible=False)
                    break

                progress.update(task, completed=completed_filezise(), visible=True)
                sleep(0.25)

        # 开启多线程
        pool = ThreadPoolExecutor()
        pool.submit(merge_segments)
        pool.submit(merge_progress)
        pool.shutdown()



from typing import List, Union, Optional

from pyxk.requests import Session
from pyxk.requests.sessions import (
    Url,
    Method,
    Params,
    Auth,
    Cert,
    Json,
    Hooks,
    Files,
    Headers,
    Cookies,
    Proxies,
    Data,
    Response
)


def request(
    method: Method,
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """request

    :param method: 请求方法 'GET', 'POST', 'HEAD', 'OPTIONS', 'DELETE', 'PUT', 'PATCH'
    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def get(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """get

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='GET',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def post(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """post

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='POST',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def head(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = False,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """head

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='HEAD',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def delete(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """delete

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='DELETE',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def put(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """put

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='PUT',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def patch(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """patch

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='PATCH',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def options(
    url: Url,
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = None,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
) -> Response:
    """options

    :param url: url
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.request(
            method='OPTIONS',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
        )


def downloader(
    url: Url,
    method: Method = 'GET',
    *,
    params: Params = None,
    data: Data = None,
    headers: Headers = None,
    cookies: Cookies = None,
    files: Files = None,
    auth: Auth = None,
    timeout: Optional[Union[int, float]] = 5,
    allow_redirects: bool = True,
    proxies: Proxies = None,
    hooks: Hooks = None,
    stream: Optional[bool] = True,
    verify: Optional[Union[bool, str]] = None,
    cert: Cert = None,
    json: Json = None,
    show_status: bool = True,
    max_retries: Optional[int] = 10,
    status_force_list: Optional[List[int]] = None,
    method_write_list: Optional[List[str]] = None,
    user_agent: Optional[str] = None,
    output: Optional[str] = None,
    threads: Optional[int] = None,
    restore: bool = False,
    transient: bool = False
) -> Response:
    """request - 下载器

    :param url: url
    :param method: 请求方法 'GET', 'POST', 'HEAD', 'OPTIONS', 'DELETE', 'PUT', 'PATCH'
    :param params: params
    :param data: data
    :param headers: 请求头
    :param cookies: cookies
    :param files: files
    :param auth: auth
    :param timeout: 超时时间
    :param allow_redirects: 请求重定向
    :param proxies: 请求代理
    :param hooks: hooks - {'response': [callable, ...]}
    :param stream: 流式相应 大文件请求
    :param verify: verify ssl
    :param cert: 请求证书
    :param json: json
    :param show_status: 请求显示 rich.status
    :param max_retries: 最大重试次数
    :param status_force_list: 状态码重试列表
    :param method_write_list: 请求方式白名单
    :param user_agent: User Agent
    :param output: 文件输出路径
    :param threads: 下载线程数量
    :param restore: 文件续传
    :param transient: 转瞬即逝
    :return: requests.Response
    """
    with Session(user_agent=user_agent) as session:
        return session.downloader(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list,
            output=output,
            threads=threads,
            restore=restore,
            transient=transient
        )



import os
from typing import Optional, Union
from multidict import CIMultiDict

import m3u8

from pyxk.m3u8._initial import _M3U8Init
from pyxk.m3u8.downloader import Downloader
from pyxk import (
    chardet,
    normpath,
    LazyLoader,
    human_playtime,
    read_file_by_generator
)

_box = LazyLoader("_box", globals(), "rich.box")
_live = LazyLoader("_live", globals(), "rich.live")
_panel = LazyLoader("_panel", globals(), "rich.panel")
_table = LazyLoader("_table", globals(), "rich.table")


__all__ = ['M3u8Parser', 'load_url', 'load_content']


class M3u8Parser(_M3U8Init):

    def load_content(
        self,
        content: str,
        url: Optional[str] = None,
        output: Optional[str] = None,
        verify: Optional[bool] = None,
        headers: Union[dict, CIMultiDict, None] = None,
        user_agent: Optional[str] = None,
        reload: Optional[bool] = None,
        reserve: Optional[bool] = None,
        limit: Optional[int] = None
    ):
        """load m3u8 url

        :param content: m3u8 content
        :param url: m3u8 url
        :param output: 文件输出路径
        :param verify: verify ssl
        :param headers: Headers
        :param user_agent: User agent
        :param reload: 重载m3u8文件
        :param reserve: 下载完成保留m3u8文件
        :param limit: 异步下载limit
        """
        # 文件路径
        self.output = output

        # 初始化数据
        self._initialization(
            verify=verify,
            headers=headers,
            user_agent=user_agent,
            reload=reload,
            reserve=reserve,
            limit=limit
        )

        # 无效 m3u8 content
        if not isinstance(content, str) or not content:
            raise TypeError('Invalid m3u8 content')

        # 查看是否是本地文件
        file = normpath(os.path.join(self.store, content))
        if not self._is_m3u8(content) and os.path.isfile(file):
            _content = b''

            for index, chunk in enumerate(read_file_by_generator(file, 'rb')):
                # 检查是否是m3u8内容
                if index == 0 and not self._is_m3u8(chunk):
                    break
                _content += chunk
            content = _content.decode(
                chardet(_content).encoding or 'utf-8'
            )

        # 解析m3u8
        self.start_parse(content, url)

    def load_url(
        self,
        url: str,
        output: Optional[str] = None,
        verify: Optional[bool] = None,
        headers: Union[dict, CIMultiDict, None] = None,
        user_agent: Optional[str] = None,
        reload: Optional[bool] = None,
        reserve: Optional[bool] = None,
        limit: Optional[int] = None
    ):
        """load m3u8 url

        :param url: m3u8 url
        :param output: 文件输出路径
        :param verify: verify ssl
        :param headers: Headers
        :param user_agent: User agent
        :param reload: 重载m3u8文件
        :param reserve: 下载完成保留m3u8文件
        :param limit: 异步下载limit
        """
        # 文件路径
        self.output = output

        # 初始化数据
        self._initialization(
            verify=verify,
            headers=headers,
            user_agent=user_agent,
            reload=reload,
            reserve=reserve,
            limit=limit
        )
        content = self.get_m3u8_content(url)

        # 无效m3u8数据
        if not content:
            self.console.print(
                f'[red b]Invalid m3u8 url[/]: <{url}>'
            )
            return

        # 解析m3u8
        self.start_parse(content=content, url=url)

    def start_parse(self, content: str, url: Optional[str] = None):
        """开始解析m3u8数据

        :param content: m3u8内容
        :param url: m3u8 url
        """
        m3u8obj = m3u8.loads(content=content, uri=url)
        m3u8obj, new_url = self._parse_playlists(m3u8obj, url)
        m3u8keys = self._parse_m3u8keys(m3u8obj)
        segments, durations = self._parse_segments(m3u8obj, new_url)

        # 无效m3u8数据
        if not segments:
            self.console.print(f'[red b]Invalid m3u8 content[/], url: {url}')
            return

        # 可视化解析结果
        table = _table.Table(show_header=False, box=_box.SIMPLE_HEAD)

        table.add_column(overflow='fold')
        table.add_row(f'[blue u]{url}[/]')
        table.add_section()

        m3u8_info = [
            f'[yellow b]maximum[/]: {len(segments)}',
            f'[yellow b]durations[/]: {human_playtime(durations)}',
            f'[yellow b]output[/]: {self.output}'
        ]
        for row in m3u8_info:
            table.add_row(row)

        panel = _panel.Panel(
            table,
            box=_box.ASCII2,
            border_style='bright_blue',
            title='[red]M3U8 Downloader[/]',
            title_align='center',
            subtitle=f'[dim i]limit: {self.limit}[/]',
            subtitle_align='right',
            expand=False,
        )

        # 没有 output 展示解析结果即可
        if not self.output:
            self.console.print(panel)
            return

        # 创建segments文件夹
        os.makedirs(self._temp_segments, exist_ok=True)

        # 开启下载
        with _live.Live(panel, console=self.console):
            Downloader.run(
                start_urls=segments,
                m3u8keys=m3u8keys,
                semaphore=self.limit,
                headers=self.headers,
                user_agent=self.user_agent,
                verify=self._verify,
                output=self.output,
                table=table,
                console=self.console,
                temp=self._temp_segments,
            )

        # 删除m3u8文件
        if not os.path.isfile(self.output):
            self._reserve = False

        if self._reserve is True:
            import shutil
            shutil.rmtree(self.temp)
        # 清空请求历史
        self.history.clear()

    def _parse_segments(self, m3u8obj, url):
        """解析 m3u8 segments"""
        # 无效的 m3u8 文件
        if not m3u8obj.is_endlist:
            return None, None

        # all segments
        segments, durations = [], 0

        for index, segment in enumerate(m3u8obj.segments):
            # segments 绝对路径
            segment.uri = segment.absolute_uri

            # segment key
            key = segment.key.uri if segment.key else None

            # 保存 segment
            item = (
                segment.uri,
                {
                    'file': normpath(
                        os.path.join(self._temp_segments, str(index)+'.ts')
                    ) if self.output else None,
                    'key': key
                }
            )
            segments.append(item)
            # segment 时间累加
            durations += segment.duration

        # 保存 m3u8 文件
        self.sava_m3u8_content(url, m3u8obj.dumps())
        return segments, durations

    def _parse_m3u8keys(self, m3u8obj):
        """解析 m3u8 keys"""
        keys = {}

        for key in m3u8obj.keys:
            # 无效key
            if not key:
                continue

            key.uri = key.absolute_uri
            # 获取 密钥 和 偏移量
            secret = self.get_m3u8_content(key.uri, True)[:16]
            iv = key.iv.removeprefix('0x')[:16].encode() if key.iv else secret
            # 保存key
            keys[key.uri] = {'key': secret, 'iv': iv}

            # 保存密钥到本地
            self.sava_m3u8_content(key.uri, secret, True)

        return keys

    def _parse_playlists(self, m3u8obj, url=None):
        """解析 m3u8 playlists"""

        # 没有 playlists
        if not m3u8obj.is_variant:
            return m3u8obj, url

        def sorted_playlists(playlist):
            """对 playlists 进行排序"""
            playlist.uri = playlist.absolute_uri
            return playlist.stream_info.bandwidth

        # 对 playlists 进行排序
        playlists = sorted(m3u8obj.playlists, key=sorted_playlists)

        # 保存m3u8文件
        self.sava_m3u8_content(url, m3u8obj.dumps())

        # 获取带宽最大的 playlist
        new_url = playlists[-1].uri
        new_m3u8obj = m3u8.loads(self.get_m3u8_content(new_url), new_url)

        return self._parse_playlists(new_m3u8obj, new_url)


def load_content(
    content: str,
    url: Optional[str] = None,
    output: Optional[str] = None,
    verify: Optional[bool] = None,
    headers: Union[dict, CIMultiDict, None] = None,
    user_agent: Optional[str] = None,
    reload: Optional[bool] = None,
    reserve: Optional[bool] = None,
    limit: Optional[int] = None
):
    """load m3u8 url

    :param content: m3u8 content
    :param url: m3u8 url
    :param output: 文件输出路径
    :param verify: verify ssl
    :param headers: Headers
    :param user_agent: User agent
    :param reload: 重载m3u8文件
    :param reserve: 下载完成保留m3u8文件
    :param limit: 异步下载limit
    """
    m3u8parse = M3u8Parser()
    m3u8parse.load_content(
        content=content,
        url=url,
        output=output,
        verify=verify,
        headers=headers,
        user_agent=user_agent,
        reload=reload,
        reserve=reserve,
        limit=limit
    )


def load_url(
    url: str,
    output: Optional[str] = None,
    verify: Optional[bool] = None,
    headers: Union[dict, CIMultiDict, None] = None,
    user_agent: Optional[str] = None,
    reload: Optional[bool] = None,
    reserve: Optional[bool] = None,
    limit: Optional[int] = None
):
    """load m3u8 url

    :param url: m3u8 url
    :param output: 文件输出路径
    :param verify: verify ssl
    :param headers: Headers
    :param user_agent: User agent
    :param reload: 重载m3u8文件
    :param reserve: 下载完成保留m3u8文件
    :param limit: 异步下载limit
    """
    m3u8parse = M3u8Parser()
    m3u8parse.load_url(
        url=url,
        output=output,
        verify=verify,
        headers=headers,
        user_agent=user_agent,
        reload=reload,
        reserve=reserve,
        limit=limit
    )



import click
from pyxk.requests import downloader
from pyxk.utils import LazyLoader, get_user_agent

_json = LazyLoader("_json", globals(), "json")


def validate_json(ctx, param, value):
    if value is None:
        return None
    try:
        return _json.loads(value)
    except _json.JSONDecodeError:  # pragma: no cover
        raise click.BadParameter("Not valid JSON")


def validate_auth(ctx, param, value):
    if value == (None, None):
        return None
    username, password = value
    if password == "-":  # pragma: no cover
        password = click.prompt("Auth Password", hide_input=True)
    return username, password


@click.command()
@click.argument("url", type=str)
@click.option(
    "--method",
    "-m",
    "method",
    type=str,
    help=(
        "Request method (GET, POST, PUT, PATCH, DELETE, OPTIONS, HEAD)."
    )
)
@click.option(
    "--params",
    "-p",
    "params",
    type=(str, str),
    multiple=True,
    help="Request parameters.",
)
@click.option(
    "--data",
    "-d",
    "data",
    type=(str, str),
    multiple=True,
    help="Form data.",
)
@click.option(
    "--files",
    "-f",
    "files",
    type=(str, click.File(mode="rb")),
    multiple=True,
    help="Form files.",
)
@click.option(
    "--json",
    "-j",
    "json",
    type=str,
    callback=validate_json,
    help="Json data.",
)
@click.option(
    "--headers",
    "-h",
    "headers",
    type=(str, str),
    multiple=True,
    help="Http headers.",
)
@click.option(
    "--cookies",
    "cookies",
    type=(str, str),
    multiple=True,
    help="Request cookies.",
)
@click.option(
    "--auth",
    "auth",
    type=(str, str),
    default=(None, None),
    callback=validate_auth,
    help="Username and password"
)
@click.option(
    "--proxies",
    "proxies",
    type=str,
    default=None,
    help="Http or https proxies.",
)
@click.option(
    "--timeout",
    "-t",
    "timeout",
    type=float,
    default=5.0,
    help="Http connection timeout(default: 5.0)"
)
@click.option(
    "--no-allow-redirects",
    "allow_redirects",
    is_flag=True,
    default=True,
    help="Disable automatically follow redirects.",
)
@click.option(
    "--no-verify",
    "verify",
    is_flag=True,
    default=True,
    help="Disable SSL verification.",
)
@click.option(
    "--output",
    "-o",
    "output",
    type=str,
    default=None,
    help="Save to file"
)
@click.option(
    "--restore",
    "-r",
    "restore",
    is_flag=True,
    default=False,
    help="Restore connection"
)
@click.option(
    "--threads",
    "threads",
    type=int,
    default=1,
    help="Multi threaded download"
)
@click.option(
    "-ua",
    "--user-agent",
    "user_agent",
    type=str,
    default=None,
    help="User-Agent"
)
def main(
    method,
    url,
    params,
    headers,
    data,
    json,
    files,
    cookies,
    auth,
    proxies,
    timeout,
    allow_redirects,
    verify,
    output,
    restore,
    threads,
    user_agent
):
    if not method:
        method = "POST" if data or files or json else "GET"

    downloader(
        method=method,
        url=url,
        params=dict(params) or None,
        headers=dict(headers) or None,
        json=json or None,
        files=dict(files) or None,
        cookies=dict(cookies) or None,
        proxies=proxies,
        timeout=timeout,
        allow_redirects=allow_redirects,
        verify=verify,
        auth=auth,
        output=output,
        restore=restore,
        threads=threads,
        user_agent=get_user_agent(user_agent),
        show_status=False
    )



import io
import logging
from types import ModuleType
from importlib import import_module
from collections import namedtuple
from typing import IO, Any, Union, Optional


__all__ = [
    'LazyLoader', 'runtime', 'runtime_coro', 'make_open',
    'get_user_agent', 'default_headers', 'md5', 'hash256', 'normpath',
    'is_base64_from_regex', 'base64_conversion_bytes', 'rename_file', 'rename_folder', 'rename',
    'str_conversion_digit', 'human_playtime', 'pycode_conversion_lazy_loader', 'pyfile_conversion_lazy_loader',
    'units_conversion_from_byte', 'download_progress', 'tasks_progress', 'chardet', 'read_file_by_generator'
]


class LazyLoader(ModuleType):
    """模块延迟加载"""

    def __init__(
        self,
        local_name: str,
        parent_module_globals: dict,
        name: Optional[str] = None,
        warning: Optional[str] = None
    ):
        """模块延迟加载 init

        :param local_name: 父模块引用名称
        :param parent_module_globals: 父模块全局变量
        :param name: 导入模块名称
        :param warning: 警告信息
        """
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        self._warning = warning
        super().__init__(name or local_name)

    def _load(self):
        """加载模块并将其插入父模块的全局变量中"""
        # 导入模块
        module = import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        # 如果指定了警告，则发出警告
        if self._warning:
            logging.warning(self._warning)
            # 确保只警告一次
            self._warning = None

        # 将模块的方法和变量注册到当前对象下
        self.__dict__.update(module.__dict__)
        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


os = LazyLoader('os', globals(), 'os')
re = LazyLoader('re', globals(), 're')
time = LazyLoader('time', globals(), 'time')
base64 = LazyLoader('base64', globals(), 'base64')
hashlib = LazyLoader('hashlib', globals(), 'hashlib')
difflib = LazyLoader('difflib', globals(), 'difflib')
warnings = LazyLoader('warnings', globals(), 'warnings')
functools = LazyLoader('functools', globals(), 'functools')
itertools = LazyLoader('itertools', globals(), 'itertools')
collections = LazyLoader('collections', globals(), 'collections')
_console = LazyLoader('_console', globals(), 'rich.console')


def runtime(func):
    """装饰器: 计算函数运行时间"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        con = _console.Console()
        start_time = time.perf_counter()
        try:
            result = func(*args, **kwargs)
        finally:
            end_time = time.perf_counter()
            con.print(f'{func.__name__!r} [magenta]running time[/]: {end_time - start_time}')
        return result

    return wrapper


def runtime_coro(func):
    """装饰器: 计算异步函数运行时间"""

    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        con = _console.Console()
        start_time = time.perf_counter()
        try:
            result = await func(*args, **kwargs)
        finally:
            end_time = time.perf_counter()
            con.print(f'{func.__name__!r} [magenta]running time[/]: {end_time - start_time}')
        return result

    return wrapper


def _make_open(func):
    """内置方法open装饰器 - 文件模式 w/a 下，创建不存在的目录"""

    @functools.wraps(func)
    def wrapper(
        file, mode='r', buffering=-1, encoding=None, errors=None, newline=None, closefd=True, opener=None
    ) -> IO:
        if not isinstance(mode, str):
            raise TypeError(f"{func.__name__}() argument 'mode' must be str, not {type(mode).__name__!r}")

        # collections.Counter 统计可迭代对象 每项出现的次数 & itertools.product 求多个可迭代对象的笛卡尔积
        mode_list = [collections.Counter(i + j) for i, j in itertools.product('wa', ('b', 'b+', '', '+'))]
        if collections.Counter(mode) in mode_list:
            os.makedirs(os.path.dirname(file), exist_ok=True)

        # 二进制模式下 encoding=None
        if mode.find('b') != -1 and encoding is not None:
            logging.warning("binary mode doesn't take an encoding argument")
            encoding = None

        return func(file, mode, buffering, encoding, errors, newline, closefd, opener)
    return wrapper


make_open = _make_open(io.open)


def get_user_agent(ua: Optional[str] = 'android', overwrite: bool = False) -> Union[str, Any]:
    """获取 UserAgent，默认 Android

    :param ua: 模糊查找内置字典User-Agent (android, windows, mac, iphone, ipad, symbian, apad)
    :param overwrite: 若为True, 直接返回  User-Agent
    :return: str
    """
    # 重写 UserAgent
    if overwrite:
        return str(ua)

    # UserAgent 全部字典
    user_agent_dict = {
        'android': 'Mozilla/5.0 (Linux; Android 11; Pixel 5) AppleWebKit/537.36 (KHTML, like Gecko) '
                   'Chrome/90.0.4430.91 Mobile Safari/537.36',
        'windows': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) '
                   'Chrome/86.0.4240.198 Safari/537.36',
        'mac': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) '
               'Chrome/86.0.4240.198 Safari/537.36',
        'iphone': 'Mozilla/5.0 (iPhone; CPU iPhone OS 13_2_3 like Mac OS X) '
                  'AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Mobile/15E148 Safari/604.1',
        'ipad': 'Mozilla/5.0 (iPad; CPU OS 13_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) '
                'CriOS/87.0.4280.77 Mobile/15E148 Safari/604.1',
        'symbian': 'Mozilla/5.0 (Symbian/3; Series60/5.2 NokiaN8-00/012.002; Profile/MIDP-2.1 Configuration/CLDC-1.1 ) '
                   'AppleWebKit/533.4 (KHTML, like Gecko) NokiaBrowser/7.3.0 Mobile Safari/533.4 3gpp-gba',
        'apad': 'Mozilla/5.0 (Linux; Android 11; Phh-Treble vanilla Build/RQ3A.211001.001;) '
                'AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/90.0.4430.91 Safari/537.36',
    }

    if not isinstance(ua, str):
        return user_agent_dict['android']

    # 模糊查找 UserAgent
    ua = difflib.get_close_matches(ua.lower(), user_agent_dict.keys(), 1)
    if not ua:
        return user_agent_dict['android']
    return user_agent_dict[ua[0]]


def default_headers(ua: Optional[str] = 'android'):
    """Headers

    :param ua: user-agent
    :return: dict -> {'User-Agent': str}
    """
    return {'User-Agent': get_user_agent(ua)}


def md5(plaintext: Union[str, bytes], encoding: str = 'utf-8', default: Any = None):
    """MD5加密

    :param plaintext: 需加密明文
    :param encoding: plaintext编码
    :param default: ciphertext默认值
    :return: tuple -> (plaintext: 'plaintext', ciphertext: 'ciphertext')
    """

    _tuple = namedtuple('MD5NameTuple', 'ciphertext,plaintext')
    result = _tuple(default, plaintext)

    if not isinstance(plaintext, (str, bytes)):
        return result

    # plaintext type str
    if isinstance(plaintext, str):
        plaintext = plaintext.encode(encoding=encoding)

    # md5加密
    result = result._replace(
        ciphertext=hashlib.md5(plaintext).hexdigest()
    )
    return result


def hash256(plaintext: Union[str, bytes], encoding: str = 'utf-8', default: Any = None):
    """HASH_256加密

    :param plaintext: 需加密明文
    :param encoding: plaintext编码
    :param default: ciphertext默认值
    :return: tuple -> (plaintext: 'plaintext', ciphertext: 'ciphertext')
    """
    _tuple = namedtuple('Hash256NameTuple', 'ciphertext,plaintext')
    result = _tuple(default, plaintext)

    if not isinstance(plaintext, (str, bytes)):
        return result

    # plaintext type str
    if isinstance(plaintext, str):
        plaintext = plaintext.encode(encoding=encoding)

    # hash256加密
    result = result._replace(
        ciphertext=hashlib.sha256(plaintext).hexdigest()
    )
    return result


def is_base64_from_regex(data: Union[str, bytes]) -> bool:
    """判断是否为base64数据类型

    :param data: 需检测的数据
    :return: bool
    """
    # base64 数据类型 正则表达式判断
    if isinstance(data, bytes):
        pattern_from_bytes = re.compile(rb'^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$')
        return bool(pattern_from_bytes.match(data))

    # base64 数据类型 正则表达式判断
    if isinstance(data, str):
        pattern_from_string = re.compile(r'^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)?$')
        return bool(pattern_from_string.match(data))
    # str 或 bytes 以外类型返回 False
    return False


def base64_conversion_bytes(data: Union[str, bytes], encoding: str = 'utf-8', default: Any = None):
    """base64数据类型 转化为bytes

    :param data: data: data
    :param encoding: data数据编码
    :param default: 返回默认值
    :return: tuple -> (result: 'result', is_base64: 'is_base64')
    """
    _tuple = namedtuple('Base64NameTuple', 'result,is_base64')
    result = _tuple(default, False)

    if (
        not isinstance(data, (str, bytes))
        or not is_base64_from_regex(data)
    ):
        return result

    if isinstance(data, str):
        data = data.encode(encoding)


    result = result._replace(
        result=base64.b64decode(data),
        is_base64=True
    )

    return result


def normpath(_path) -> str:
    """规范文件路径"""
    if not isinstance(_path, str):
        raise TypeError(f'_path must be a str, got {type(_path).__name__!r}')

    return os.path.normpath(_path.strip()).replace('\\', '/')


def rename_file(file: str, suffix: Optional[str] = None, create: bool = False):
    """重命名本地存在的文件

    :param file: 文件路径
    :param suffix: 文件后缀名
    :param create: 创建文件
    :return: tuple -> (result, dirname, basename)
    """
    # file 绝对路径
    file = normpath(os.path.abspath(file))
    if not os.path.basename(file):
        raise ValueError('file missing basename')

    # file 后缀名
    if suffix and isinstance(suffix, str):
        suffix = '.' + suffix.removeprefix('.')
    else:
        basename = os.path.basename(file).rsplit('.', 1)
        if len(basename) == 2 and 2 <= len(basename[1]) <= 8:
            suffix = '.' + basename[1].removeprefix('.')
        else:
            suffix = ''

    # 拼接完整file
    if not file.endswith(suffix):
        file += suffix

    # 重命名文件
    def _rename(f):
        """rename a file"""
        _tuple = namedtuple('RenameFileNameTuple', 'result,dirname,basename')

        if create:
            from pathlib import Path
            Path(f).touch()
        return _tuple(f, *os.path.split(f))

    if not os.path.isfile(file):
        return _rename(file)

    for index in itertools.count(1):
        newfile = file.removesuffix(suffix) + f'.{index}{suffix}'
        if not os.path.isfile(newfile):
            return _rename(newfile)


rename = rename_file


def rename_folder(folder: str, create: bool = False):
    """重命名本地存在的文件夹

    :param folder: 文件夹路径
    :param create: 创建文件夹
    :return: Tuple(result, dirname, basename)
    """
    # 获取folder绝对路径
    folder = normpath(os.path.abspath(folder))

    def get_folder(f):
        """获取重命名文件数据"""
        _tuple = namedtuple('RenameFolderNameTuple', 'result,dirname,basename')

        if create:
            os.makedirs(f, exist_ok=True)
        return _tuple(f, *os.path.split(f))

    if not os.path.isdir(folder):
        return get_folder(folder)

    # 重命名folder
    for index in itertools.count(1):
        new_folder = folder + f".{index}"
        if not os.path.isdir(new_folder):
            return get_folder(new_folder)


def str_conversion_digit(target: Union[str, int, float], default: Any = None):
    """字符串转换为数字

    :param target: 需要转换的目标
    :param default: 返回默认值
    :return: tuple -> (result: digits, raw, is_digit: bool)
    """
    _tuple = namedtuple('StrConversionDigitNameTuple', 'result,raw,is_digit')
    result = _tuple(default, target, False)

    # target type = `int` or `float`
    if isinstance(target, (int, float)):
        return result._replace(
            is_digit=True, result=target
        )

    # target type != `str`
    if not isinstance(target, str):
        return result

    # 判断字符串是否为数字
    def is_digits(string):
        pattern = re.match(r"^(?P<symbol>-)?(?P<int>\d+)(?P<float>\.\d+)?$", string)
        if not pattern:
            return {"is_digits": False, "type": lambda x: x}
        return {"is_digits": True, "type": float if pattern.groupdict()["float"] is not None else int}

    # 转换为数字
    _result = is_digits(target)
    if _result["is_digits"]:
        result = result._replace(
            is_digit=True, result=_result["type"](target)
        )
    return result


def human_playtime(playtime: Union[str, int, float], default: Any = None):
    """人类直观时间展示

    :param playtime: 传入一个时间(秒), 返回人类能理解的时间格式
    :param default: 返回默认值
    :return: str
    """
    digit = str_conversion_digit(playtime)

    if not digit.is_digit:
        return default

    playtime = digit.result
    symbol, playtime = "-" if playtime < 0 else "", round(abs(playtime))

    hour, second = divmod(playtime, 3600)
    minute, second = divmod(second, 60)

    return f"{symbol}{hour}:{minute:0>2}:{second:0>2}"


def pycode_conversion_lazy_loader(string: str):
    """python代码中的导入模块转换为懒加载

    :param string: python代码
    :return: str
    """
    def repl_string(match):
        """替换懒加载"""
        match_dict, import_name, alias = match.groupdict(), None, None
        repl = '{alias} = LazyLoader("{alias}", globals(), "{import_name}")'

        # from导入
        if match_dict["from_name"]:
            import_name = f'{match_dict["from_name"]}.{match_dict["from_import_name"]}'
            alias = match_dict["from_import_name"]
        else:
            import_name = f'{match_dict["import_name"]}'
            alias = match_dict["import_name"]

        # alias
        if match_dict["from_import_alias"]:
            alias = match_dict["from_import_alias"]
        elif match_dict["import_alias"]:
            alias = match_dict["import_alias"]
        return repl.format(alias=alias, import_name=import_name)

    # 替换懒加载
    pattern = [
        r'^from\s+?(?P<from_name>\S+)\s+?import\s+?(?P<from_import_name>\S+)'
        r'\s*?(as\s+?(?P<from_import_alias>\S+))?$',
        r'^import\s+?(?P<import_name>\S+)(\s+?as\s+(?P<import_alias>\S+?))?\s*?$',
    ]
    return re.sub(pattern='|'.join(pattern), repl=repl_string, string=string, flags=re.M)


def pyfile_conversion_lazy_loader(read_file: str, write_file: str, encoding: Optional[str] = None):
    """python文件中的导入模块转换为懒加载

    :param read_file: 读取python代码文件
    :param write_file: 写入转换后的python代码文件
    :param encoding: 文件编码
    """
    with open(read_file, "r", encoding=encoding) as read_file_obj:
        content = pycode_conversion_lazy_loader(read_file_obj.read())
    with open(write_file, "w", encoding=encoding) as write_file_obj:
        write_file_obj.write(content)


def units_conversion_from_byte(target: Union[str, int, float], default: Any = None):
    """字节单位自动换算

    :param target: 换算目标(Bytes)
    :param default: 返回默认值
    :return: tuple -> (result, raw, units)
    """
    _tuple = namedtuple('UnitsConversionNameTuple', 'result,raw,units')
    result = _tuple(default, target, None)
    _temp = str_conversion_digit(target)

    if not _temp.is_digit:
        return result

    target = abs(_temp.result)
    target_units = 'Bytes'
    del _temp

    units_dict = {
        'Bytes': 1, 'KB': 1024, 'MB': 1024,
        'GB': 1024, 'TB': 1024, 'PB': 1024,
        'EB': 1024, 'ZB': 1024, 'YB': 1024, 'BB': 1024
    }

    for units, rate in units_dict.items():
        if target >= rate:
            target, target_units = target/rate, units
            continue
        break

    result = result._replace(
        units=target_units,
        result= f'{round(target, 2)}{target_units}'

    )
    return result


def chardet(byte_str: bytes, **kwargs):
    """字符编码判断

    :param byte_str: bytes数据
    :param kwargs: kwargs
    :return: tuple -> (encoding: str, language: str, confidence: float)
    """
    try:
        import chardet as _chardet
    except ImportError:
        import charset_normalizer as _chardet

    _tuple = namedtuple('ChardetNameTuple', 'encoding,language,confidence')
    result = _chardet.detect(byte_str=byte_str, **kwargs)

    return _tuple(
        encoding=result.get('encoding'),
        language=result.get('language'),
        confidence=result.get('confidence')
    )



def download_progress(
    *,
    transient: bool = False,
    console: object = None,
    text_column: Optional[str] = '[progress.description]{task.description}',
    task_progress_column: Optional[str] = '[progress.percentage]{task.percentage:>6.2f}%',
    show_transfer_speed: bool = True
):
    """rich.progress.Progress - 下载进度条

    :param transient: 转瞬即逝
    :param console: rich.console.Console
    :param text_column: text column 文本格式 [progress.description]{task.description}
    :param task_progress_column: 任务进度格式 [progress.percentage]{task.percentage:>6.2f}%
    :param show_transfer_speed: 显示任务下载速度
    :return: rich.progress.Progress default: True
    """
    from rich import progress as _progress
    columns = [
        _progress.TextColumn(text_column),
        _progress.TaskProgressColumn(task_progress_column),
        _progress.BarColumn(),
        _progress.DownloadColumn(),
        _progress.TransferSpeedColumn(),
        _progress.TimeElapsedColumn()
    ]
    # 不显示下载速度
    if not show_transfer_speed:
        columns.pop(-2)
    return _progress.Progress(*columns, transient=transient, console=console)


def tasks_progress(
    *,
    transient: bool = False,
    console: object = None,
    text_column: Optional[str] = '[progress.description]{task.description}',
    task_progress_column: Optional[str] = '[progress.percentage]{task.percentage:>6.2f}%',
    task_progress_column_2: Optional[str] = '[cyan]{task.completed}/{task.total}[/]',
    show_transfer_speed: bool = False
):
    """rich.progress.Progress - 任务进度条

    :param transient: 转瞬即逝
    :param console: rich.console.Console
    :param text_column: text column 文本格式 [progress.description]{task.description}
    :param task_progress_column: 任务进度格式 [progress.percentage]{task.percentage:>6.2f}%
    :param task_progress_column_2: 任务进度格式 [cyan]{task.completed}/{task.total}[/]
    :param show_transfer_speed: 显示任务下载速度 default: False
    :return:
    """
    from rich import progress as _progress
    columns = [
        _progress.TextColumn(text_column),
        _progress.TaskProgressColumn(task_progress_column),
        _progress.BarColumn(),
        _progress.TaskProgressColumn(task_progress_column_2),
        _progress.TimeElapsedColumn(),
    ]
    # 显示下载速度
    if show_transfer_speed:
        columns.insert(-1, _progress.TransferSpeedColumn())
    return _progress.Progress(*columns, transient=transient, console=console)


def read_file_by_generator(file: str, mode: str = 'r', *, chunk_size: int = 1024, **kwargs):
    """基于生成器的文件读取

    :param file: file path
    :param mode: file open mode
    :param chunk_size: chunk size
    :param kwargs: open arguments
    """
    with open(file, mode, **kwargs) as read_file_obj:
        while True:
            chunk = read_file_obj.read(chunk_size)
            if not chunk:
                return
            yield chunk



import asyncio
from types import MethodType
from typing import (
    Any, List, Tuple, Union, Optional, Mapping, Callable
)

from multidict import CIMultiDict
from pyxk.aclient.typedef import (
    URL, Session, Response, EventLoop, CIMDict, Timeout, URLS
)
from pyxk.utils import (
    LazyLoader, get_user_agent, default_headers
)

copy = LazyLoader('copy', globals(), 'copy')
yarl = LazyLoader('yarl', globals(), 'yarl')
urllib = LazyLoader('urllib', globals(), 'urllib.parse')
aiohttp = LazyLoader('aiohttp', globals(), 'aiohttp')
logging = LazyLoader('logging', globals(), 'logging')
selector = LazyLoader('selector', globals(), 'parsel.selector')
aiohttp_exceptions = LazyLoader('aiohttp_exceptions', globals(), 'aiohttp.client_exceptions')


__all__ = ['Client', 'Response']


class Client:
    """异步下载器

    explain:
    from pyxk import Client, Response

    class Download(Client):
        start_urls = ['http://www.baidu.com' for _ in range(10)]

        async def parse(self, response: Response, **kwargs):
            print(response.url)

    Download.run()
    """

    limit: int = 100
    timeout: Timeout = None
    verify: bool = True
    warning: bool = False
    headers: Union[dict, CIMDict] = CIMultiDict(default_headers())
    semaphore: Union[int, asyncio.Semaphore] = 32
    user_agent: Optional[str] = None
    req_kwargs: dict = {}
    start_urls: Union[List[str], List[Tuple[str, dict]]] = []
    async_sleep: Union[int, float] = 0
    max_retries: int = 15
    aiohttp_kwargs: dict = {}
    status_retry_list: List[int] = []
    status_error_list: List[int] = []
    until_request_succeed: Union[bool, List[int]] = True
    ATTRS = {
        'timeout': 10,
        'semaphore': 16,
        'user_agent': get_user_agent(),
    }

    def __init__(self, *, base_url: URL = None, **kwargs):
        """异步下载器

        :param limit: aiohttp 并发控制
        :param timeout: 请求超时时间
        :param verify: ssl验证
        :param warning: 警告信息
        :param headers: 请求头
        :param semaphore: asyncio并发控制
        :param user_agent: User-Agent
        :param req_kwargs: start_request请求参数
        :param start_urls: 请求入口urls
        :param async_sleep: 异步休眠时间
        :param max_retries: 异步请求最大重试次数
        :param aiohttp_kwargs: aiohttp.ClientSession 实例化参数
        :param status_retry_list: 请求状态码，包含在列表中的进行重新发送
        :param status_error_list: 请求状态码，包含在列表中直接抛出错误
        :param until_request_succeed: 请求状态码，直到请求响应成功(可自定义响应状态码)
        :param base_url: base_url 每个请求URL进行拼接
        """
        # 动态生成实例变量
        for key, val in kwargs.items():
            setattr(self, key, val)

        # event loop
        self._loop: EventLoop = None

        # aiohttp session
        self._session: Session = None

        # base_url
        self._base_url: URL = self.set_base_url(base_url)

    @classmethod
    def run(cls, **kwargs):
        """程序运行入口 - 应该调用该方法运行"""
        self = cls(**kwargs)

        # 创建新的 event loop
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        # 运行
        self.loop.run_until_complete(self.async_start())

        # 关闭 EventLoop
        if self.loop:
            self.loop.close()
            asyncio.set_event_loop(None)

        return self

    async def async_start(self):
        """开启异步下载器"""
        await self.open()

        # 无效 event loop
        if not isinstance(self.loop, asyncio.AbstractEventLoop):
            raise TypeError(f'Event loop must be an instance of AbstractEventLoop, not {type(self.loop).__name__!r}')

        # event loop 已经关闭
        if self.loop.is_closed():
            raise ValueError('Event loop is not closed')

        # 配置 asyncio semaphore
        if isinstance(self.semaphore, int) and self.semaphore > 0:
            self.semaphore = asyncio.Semaphore(self.semaphore)
        elif not isinstance(self.semaphore, asyncio.Semaphore):
            self.semaphore = asyncio.Semaphore(self.__class__.ATTRS['semaphore'])

        # aiohttp.Session 实例化参数
        aiohttp_kwargs = copy.deepcopy(dict(self.aiohttp_kwargs))

        # 为 aiohttp.Session 设置超时时间
        if isinstance(self.timeout, int) and self.timeout >= 0:
            aiohttp_kwargs['timeout'] = aiohttp.ClientTimeout(total=self.timeout)
        elif not aiohttp_kwargs.__contains__('timeout'):
            aiohttp_kwargs['timeout'] = aiohttp.ClientTimeout(total=self.__class__.ATTRS['timeout'])
        self.timeout = aiohttp_kwargs['timeout'].total

        # 为 aiohttp.Session 设置连接器
        if not aiohttp_kwargs.__contains__('connector'):
            if not isinstance(self.limit, int) or self.limit <= 0:
                raise ValueError(f'Limit must be a positive integer(正整数), got {self.limit!r}')

            aiohttp_kwargs['connector'] = aiohttp.TCPConnector(
                limit=self.limit, ssl=bool(self.verify), loop=self.loop
            )

        # 为 aiohttp.Session 设置 headers(user_agent)
        self.headers = CIMultiDict(self.headers)
        # user_agent
        if self.user_agent and isinstance(self.user_agent, str):
            self.headers['User-Agent'] = self.user_agent
        elif not self.headers.__contains__('User-Agent'):
            self.headers['User-Agent'] = self.__class__.ATTRS['user_agent']
        self.user_agent = self.headers.get('User-Agent')
        aiohttp_kwargs['headers'] = self.headers

        # 创建 aiohttp.ClientSession
        self._session = aiohttp.ClientSession(loop=self.loop, **aiohttp_kwargs)

        # 初始化变量
        self._init_attrs()

        # 开启请求
        try:
            await self.start()
            ret = await self.start_request()
            await self.completed(ret)
        finally:
            await self.async_close()

    async def async_close(self):
        """关闭异步下载器"""
        await self.stop()
        # 关闭 aiohttp session
        if self.session:
            await self.session.close()
        # 关闭 aiohttp session 之后调用
        await self.close()

    async def start_request(self):
        """start_urls 默认运行方法"""
        if not self.start_urls or not isinstance(self.start_urls, (list, tuple)):
            raise NotImplementedError(f'Invalid start_urls, got {self.start_urls!r}')

        return await self.gather(self.start_urls, self.parse, set_base_url=True, **dict(self.req_kwargs))

    async def request(
        self,
        url: URL,
        callback: Optional[Callable] = None,
        *,
        method: str = "GET",
        cb_kwargs: Optional[dict] = None,
        **kwargs
    ) -> Union[Response, Any]:
        """异步请求发送以及回调

        :return:
        :param url: URL
        :param callback: 响应response 回调函数(函数是异步的)
        :param method: 请求方法(default: GET)
        :param cb_kwargs: 传递给回调函数的关键字参数
        :param kwargs: 异步请求 request参数
            params
            data
            json
            cookies
            headers
            skip_auto_headers
            auth
            allow_redirects
            max_redirects
            compress
            chunked
            expect100
            raise_for_status
            read_until_eof
            proxy
            proxy_auth
            timeout
            verify_ssl
            fingerprint
            ssl_context
            ssl
            proxy_headers
            trace_request_ctx
            read_bufsize
        :return: Response, Any
        """
        status_retry_list, status_error_list = self.status_retry_list.copy(), self.status_error_list.copy()
        url, response, result, exc_from_request = self.build_url(url), None, None, None
        warning = {
            'status_retry_list': True,
            'status_error_list': True,
            'until_request_succeed': True,
            'timeout': True,
            'client': True,
            'server_disconnected': True
        }

        async with self.semaphore:
            for _ in range(1, self.max_retries+1):
                try:
                    response = result = await self.session.request(method=method, url=url, **kwargs)

                    # 直到请求成功
                    if self.until_request_succeed:
                        if response.status in self.until_request_succeed:
                            status_retry_list, status_error_list = [], []
                        else:
                            # 提示信息
                            await self._warning('until_request_succeed', url, warning, response)
                            continue

                    # 状态码错误
                    if response.status in status_error_list:
                        raise aiohttp_exceptions.ClientError(f'<Response[{response.status}] {url}>')

                    # 状态码重试
                    if response.status in status_retry_list:
                        # 提示信息
                        await self._warning('status_retry_list', url, warning, response)
                        continue

                    # 添加 aiohttp.ClientResponse 实例方法
                    add_instance_method(response)

                    # 开启回调函数
                    if callable(callback):
                        cb_kwargs = dict(cb_kwargs or {})
                        result = await callback(response, **dict(cb_kwargs or {}))
                    break

                # 请求超时 重试
                except asyncio.exceptions.TimeoutError as exc:
                    # 提示信息
                    await self._warning('timeout', url, warning, response)
                    exc_from_request = exc

                # 连接错误 重试
                except (
                    aiohttp_exceptions.ClientOSError,
                    aiohttp_exceptions.ClientPayloadError,
                    aiohttp_exceptions.ClientConnectorError,
                ) as exc:
                    # 提示信息
                    await self._warning('client', url, warning, response)
                    exc_from_request = exc

                # 服务器拒绝连接
                except aiohttp_exceptions.ServerDisconnectedError as exc:
                    # 提示信息
                    await self._warning('server_disconnected', url, warning, response)
                    exc_from_request = exc

                finally:
                    # 关闭连接
                    if response:
                        response.close()

            else:
                # 抛出错误
                if exc_from_request:
                    raise exc_from_request
                # 达到最大请求次数
                raise RuntimeError(
                    f'<Response[{response.status if response else None}] '
                    f'{url}>, max_retries: {self.max_retries}'
                )

        return result

    def _init_attrs(self):
        # 最大重试次数
        if not isinstance(self.max_retries, int) or self.max_retries <= 0:
            raise ValueError(f'max_retries must be a positive integer(正整数), got {self.max_retries!r}')

        # 状态码重试列表
        if not isinstance(self.status_retry_list, (list, tuple)):
            raise ValueError(f'status_retry_list must be a list, got {self.status_retry_list!r}')

        # 状态码错误列表
        if not isinstance(self.status_error_list, (list, tuple)):
            raise ValueError(f'status_retry_list must be a list, got {self.status_error_list!r}')

        # 直到请求成功
        if not isinstance(self.until_request_succeed, (list, tuple)):
            self.until_request_succeed = [200] if self.until_request_succeed else []

    async def _warning(self, who: str, url: str, warning: dict, response: Response):
        """打印警告信息"""
        # 状态码
        status = response.status if response else None
        # 全部警告信息
        who_list = {
            'status_retry_list':
                f'<Response[{status}] {url}> be in '
                f'status_retry_list:{self.status_retry_list!r}',
            'until_request_succeed':
                f'<Response[{status}] {url}> be in '
                f'until_request_succeed:{self.until_request_succeed!r}',
            'timeout': f'{url}> Timeout',
            'client': f'{url}> client error',
            'server_disconnected': f'{url}> server disconnected'
        }

        # 确保 每个链接每个警告类型 只会打印一次
        if self.warning and warning[who]:
            logging.warning(who_list[who])
            warning[who] = False

        await self.sleep()

    async def gather(
        self,
        urls: URLS,
        callback: Optional[Callable] = None,
        *,
        method: str = "GET",
        cb_kwargs_list: Union[List[dict], dict] = None,
        return_exceptions: bool = False,
        set_base_url: Union[bool, str] = False,
        **req_kwargs
    ) -> list:
        """发送url列表，创建异步任务 并发发送

        :param urls: Url List
        :param callback: 响应response 回调函数(函数是异步的)
        :param method: 请求方法(default: GET)
        :param cb_kwargs_list: 回调函数关键字参数列表
        :param return_exceptions: 错误传递(default: False)
        :param req_kwargs: 异步请求 request参数
        :param set_base_url: 是否设置base_url
        :return: list
        """
        if not isinstance(urls, (list, tuple)):
            raise TypeError(f'urls must be a list, not {type(urls).__name__!r}')

        length = len(urls)
        cb_kwargs_list = self._parse_gather_cb(length, cb_kwargs_list)

        for index, item in enumerate(urls):
            url, cb_kwargs = item, {}
            # 解析 url 和 回调参数
            if isinstance(item, (list, tuple)):
                if len(item) >= 2:
                    url, cb_kwargs = item[0], dict(item[1]).copy()
                elif len(item) == 1:
                    url, cb_kwargs = item[0], {}
                else:
                    raise ValueError(f'urls invalid, got {item!r}')

            # 更新数据
            urls[index] = url
            cb_kwargs.update(cb_kwargs_list[index])
            cb_kwargs_list[index] = cb_kwargs

        # 设置base_url
        if set_base_url and isinstance(set_base_url, str):
            self.base_url = set_base_url
        elif set_base_url and not self.base_url:
            self.base_url = urllib.urljoin(urls[0], '.')

        # 提交异步任务
        tasks = [
            self.request(
                url=url,
                callback=callback,
                method=method,
                cb_kwargs=cb_kwargs,
                **req_kwargs
            )
            for url, cb_kwargs in zip(urls, cb_kwargs_list)
        ]
        return await asyncio.gather(*tasks, return_exceptions=return_exceptions)

    @staticmethod
    def _parse_gather_cb(length, cb_kwargs_list):
        """解析 gather 的回调参数"""
        # 设置 cb_kwargs_list
        if isinstance(cb_kwargs_list, (list, tuple)):
            cb_kwargs_list = cb_kwargs_list[:length]
            cb_kwargs_list.extend(
                [{}] * (length - len(cb_kwargs_list))
            )

        elif isinstance(cb_kwargs_list, dict):
            cb_kwargs_list = [cb_kwargs_list] * length

        else:
            cb_kwargs_list = [{}] * length
        # 将每项转化为字典
        cb_kwargs_list = [dict(item) for item in cb_kwargs_list]
        return cb_kwargs_list

    async def parse(self, response: Response, **kwargs):
        """默认解析函数"""
        raise NotImplementedError(f'{self.__class__.__name__}.parse not implemented')

    async def completed(self, result: list):
        """运行完成结果回调函数"""

    async def open(self):
        """创建 aiohttp session 之前调用"""

    async def close(self):
        """关闭 aiohttp session 之后调用"""

    async def start(self):
        """创建 aiohttp session 之后调用"""

    async def stop(self):
        """关闭 aiohttp session 之前调用"""

    async def sleep(self, delay: Union[int, float, None] = None, result: Any = None):
        """异步休眠

        :param delay: 休眠时间
        :param result: 返回值
        :return: result
        """
        if delay is None:
            if not isinstance(self.async_sleep, (int, float)) or self.async_sleep < 0:
                raise ValueError(f'async_sleep must be a positive integer(正整数), got {self.async_sleep!r}')
            delay = self.async_sleep

        elif not isinstance(delay, (int, float)) or delay < 0:
            raise ValueError(f'delay must be a positive integer(正整数), got {delay!r}')

        return await asyncio.sleep(delay, result=result)

    def build_url(self, _url) -> URL:
        """构造完整url地址"""
        if not isinstance(_url, (str, yarl.URL)):
            return _url
        _url = yarl.URL(_url)
        if _url.is_absolute():
            return _url
        if self._base_url and isinstance(self._base_url, yarl.URL):
            return self._base_url.join(_url)
        return _url

    @staticmethod
    def set_base_url(url: URL) -> URL:
        """设置 base_url"""
        if not url or not isinstance(url, (str, yarl.URL)):
            return None
        url = yarl.URL(url)
        if not url.is_absolute():
            # 不是绝对路径
            return None
        return url

    @property
    def loop(self) -> EventLoop:
        """event loop"""
        return self._loop

    @property
    def session(self) -> Session:
        """aiohttp session"""
        return self._session

    @property
    def base_url(self) -> URL:
        """base_url"""
        return self._base_url

    @base_url.setter
    def base_url(self, _url: URL):
        self._base_url = self.set_base_url(_url)


async def xpath(
    self,
    query: str,
    text: Optional[str] = None,
    type: Optional[str] = None,
    encoding: str = "utf8",
    base_url: Optional[str] = None,
    namespaces: Optional[Mapping[str, str]] = None,
    **kwargs: Any,
):
    """aiohttp.ClientResponse - selector.xpath

    :param self: aiohttp.ClientResponse 实例
    :param query: xpath查询字符串
    :param text: str对象
    :param type: 文件类型 - "html"(default), "json", or "xml"
    :param encoding: text encoding
    :param base_url: 为文档设置URL
    :param namespaces: `namespaces` is an optional `prefix: namespace-uri` mapping (dict)
        for additional prefixes to those registered with `register_namespace(prefix, uri)`.
        Contrary to `register_namespace()`, these prefixes are not
        saved for future calls.
    :param kwargs: xpath kwargs
    """
    if not text:
        text = await self.text()
    # selector
    sel = selector.Selector(
        text=text, type=type, base_url=base_url, encoding=encoding, namespaces=namespaces
    )
    # xpath
    return sel.xpath(query=query, **kwargs)


async def css(
    self,
    query: str,
    text: Optional[str] = None,
    type: Optional[str] = None,
    encoding: str = "utf8",
    base_url: Optional[str] = None,
    namespaces: Optional[Mapping[str, str]] = None,
):
    """aiohttp.ClientResponse - selector.css

    :param self: aiohttp.ClientResponse 实例
    :param query: xpath查询字符串
    :param text: str对象
    :param type: 文件类型 - "html"(default), "json", or "xml"
    :param encoding: text encoding
    :param base_url: 为文档设置URL
    :param namespaces:
    `namespaces` is an optional `prefix: namespace-uri` mapping (dict)
    for additional prefixes to those registered with `register_namespace(prefix, uri)`.
    Contrary to `register_namespace()`, these prefixes are not
    saved for future calls.
    """
    if not text:
        text = await self.text()
    # selector
    sel = selector.Selector(
        text=text, type=type, base_url=base_url, encoding=encoding, namespaces=namespaces
    )
    # css
    return sel.css(query=query)


async def re(
    self,
    regex: str,
    text: Optional[str] = None,
    type: Optional[str] = None,
    encoding: str = "utf8",
    base_url: Optional[str] = None,
    namespaces: Optional[Mapping[str, str]] = None,
    replace_entities: bool = True,
):
    """aiohttp.ClientResponse - selector.re

    :param self: aiohttp.ClientResponse 实例
    :param regex: 编译的正则表达式 或者 字符串
    :param text: str对象
    :param type: 文件类型 - "html"(default), "json", or "xml"
    :param encoding: text encoding
    :param base_url: 为文档设置URL
    :param namespaces:
    `namespaces` is an optional `prefix: namespace-uri` mapping (dict)
    for additional prefixes to those registered with `register_namespace(prefix, uri)`.
    Contrary to `register_namespace()`, these prefixes are not
    saved for future calls.
    :param replace_entities:
    By default, character entity references are replaced by their
    corresponding character (except for ``&amp;`` and ``&lt;``).
    Passing ``replace_entities`` as ``False`` switches off these
    replacements.
    """
    if not text:
        text = await self.text()
    # selector
    sel = selector.Selector(
        text=text, type=type, base_url=base_url, encoding=encoding, namespaces=namespaces
    )
    # re
    return sel.re(regex=regex, replace_entities=replace_entities)


def urljoin(self, _url):
    """urljoin

    :param self: aiohttp.ClientResponse 实例
    :param _url: url
    :return: url
    """
    if isinstance(_url, str):
        _url = yarl.URL(_url)
    elif not isinstance(_url, yarl.URL):
        return _url
    if _url.is_absolute():
        return _url
    return self.url.join(_url)


def add_instance_method(response):
    """为异步response添加实例方法 - re, css, xpath, urljoin"""
    setattr(response, 're', MethodType(re, response))
    setattr(response, 'css', MethodType(css, response))
    setattr(response, 'xpath', MethodType(xpath, response))
    setattr(response, 'urljoin', MethodType(urljoin, response))



import setuptools


with open('README.md', 'r', encoding='utf-8') as read_file_obj:
    long_description = read_file_obj.read()

setuptools.setup(
    name='pyxk',
    version='0.6.7',
    author='kai139',
    install_requires=[
        'requests', 'pycryptodome', 'rich', 'm3u8', 'aiohttp', 'aiofiles', 'click', 'parsel'
    ],
    entry_points={
        'console_scripts': [
            'm3u8 = pyxk.m3u8.entry_point:main',
            'req = pyxk.requests.entry_point:main'
        ],
    },
    author_email='925330867@qq.com',
    description='pyxk',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
)



from typing import Union
from collections import namedtuple

from Crypto.Cipher import AES


__all__ = ['Crypto', 'encrypt', 'decrypt', 'MODES']


class Crypto:
    """AES加解密

    from pyxk.aes.cryptor import Crypto, MODES

    cipher = Crypto(
        key=b'1234567890123456',
        mode=MODES.CBC,
        iv=b'1234567890123456'
    )
    raw_text = 'Hello World'
    ciphertext = cipher.encrypt(raw_text)
    plaintext = cipher.decrypt(ciphertext)

    print('raw text:', raw_text)
    print('ciphertext:', ciphertext)
    print('plaintext:', plaintext)

    >>>
    raw text: Hello World
    ciphertext: b'\xf7\x16\x85)|,\x91\x8c\xdbd\xaef\xc3Wbu'
    plaintext: b'Hello World'
    """
    MODES = frozenset(
        (
            ('ECB', 1), ('CBC', 2), ('CFB', 3),
            ('OFB', 5), ('CTR', 6), ('OPENPGP', 7),
            ('EAX', 9), ('CCM', 8), ('SIV', 10),
            ('GCM', 11), ('OCB', 12)
        )
    )

    def __init__(
        self,
        key: Union[str, bytes],
        mode: Union[str, int] = 'CBC',
        encoding: str = 'utf-8',
        **kwargs
    ):
        """Crypto Init

        :param key: 加密/解密 密钥
        :param mode: 加密/解密 模式
        :param encoding: encoding
        :param kwargs: kwargs
        """
        self._encoding = encoding
        self._key = self._init_key(key)
        self._mode = self._init_mode(mode)
        self._cipher_attrs = kwargs

        self._cipher = None
        self._padding = b'\x00'

    def encrypt(self, plaintext: Union[str, bytes], new_cipher=True) -> bytes:
        """加密

        :param plaintext: 加密明文
        :param new_cipher: 是否重新创建cipher
        :return: bytes
        """
        # Invalid plaintext
        if not isinstance(plaintext, (str, bytes)):
            raise TypeError(f'plaintext must be a str or bytes, got {type(plaintext).__name__!r}')

        if isinstance(plaintext, str):
            plaintext = plaintext.encode(self._encoding)

        # 填充字符
        remainder = len(plaintext) % AES.block_size or AES.block_size
        plaintext += self._padding * (AES.block_size - remainder)

        # 加密
        cipher = self.new_cipher if new_cipher else self.cipher
        ciphertext = cipher.encrypt(plaintext)

        return ciphertext

    def decrypt(self, ciphertext: bytes, new_cipher=True) -> bytes:
        """解密

        :param ciphertext: 解密密文
        :param new_cipher: 是否重新创建cipher
        :return: bytes
        """
        # Invalid ciphertext
        if not isinstance(ciphertext, bytes):
            raise TypeError(f'ciphertext must be a bytes, got {type(ciphertext).__name__!r}')

        # 解密
        cipher = self.new_cipher if new_cipher else self.cipher
        plaintext = cipher.decrypt(ciphertext)

        return plaintext.rstrip(self._padding)

    def _init_key(self, key) -> bytes:
        """cipher key initialization"""
        # Invalid key
        if not isinstance(key, (str, bytes)):
            raise TypeError(f'key must be str or bytes, got {type(key).__name__!r}')

        if isinstance(key, str):
            key = key.encode(self._encoding)

        if len(key) not in AES.key_size:
            raise ValueError(f'Invalid key length {len(key)}, must length be {AES.key_size}')
        return key

    def _init_mode(self, mode) -> int:
        """cipher mode initialization"""
        return self._get_mode_value(mode)

    def _get_mode_value(self, _mode: Union[str, int]) -> int:
        """获取mode对应的int值

        :param _mode: mode
        :return: int
        """
        # Invalid mode
        if not isinstance(_mode, (int, str)):
            raise TypeError(f'mode must be a str or int, got {type(_mode).__name__!r}')

        # mode type str
        if isinstance(_mode, str):
            _mode = _mode.upper()
            if _mode not in self.all_modes:
                raise ValueError(f'Invalid mode {_mode!r}, effective modes: {", ".join(self.all_modes.keys())}')
            return self.all_modes[_mode]

        # mode type int
        if _mode not in self.all_modes.values():
            raise ValueError(
                f'Invalid mode {_mode!r}, effective modes: '
                f'{", ".join([str(i) for i in sorted(self.all_modes.values())])}'
            )
        return _mode

    def create_cipher(self):
        """创建 AES.cipher"""
        return AES.new(key=self.key, mode=self.mode, **self.attrs)

    @property
    def all_modes(self) -> dict:
        """all modes"""
        return dict(self.__class__.MODES)

    @property
    def cipher(self):
        """AES Crypto Cipher"""
        if not self._cipher:
            self._cipher = self.create_cipher()
        return self._cipher

    @property
    def new_cipher(self):
        """每次调用新建 AES Crypto Cipher"""
        self._cipher = self.create_cipher()
        return self._cipher

    @property
    def key(self):
        """AES Crypto key"""
        return self._key

    @property
    def mode(self):
        """AES Crypto mode"""
        return self._mode

    @property
    def attrs(self):
        """AES Crypto Cipher attributes"""
        return self._cipher_attrs

    @attrs.setter
    def attrs(self, value):
        self._cipher_attrs.update(dict(value))


def encrypt(
    key: Union[str, bytes],
    plaintext: Union[str, bytes],
    mode: Union[str, int] = 'CBC',
    **kwargs
) -> bytes:
    """加密

    :param key: 加密密钥
    :param plaintext: 加密明文
    :param mode: 加密模式
    :param kwargs: kwargs
    :return: bytes
    """
    cipher = Crypto(key=key, mode=mode, **kwargs)
    return cipher.encrypt(plaintext, new_cipher=True)


def decrypt(
    key: Union[str, bytes],
    ciphertext: bytes,
    mode: Union[str, int] = 'CBC',
    **kwargs
) -> bytes:
    """解密

    :param key: 解密密钥
    :param ciphertext: 解密密文
    :param mode: 解密模式
    :param kwargs: kwargs
    :return: bytes
    """
    cipher = Crypto(key=key, mode=mode, **kwargs)
    return cipher.decrypt(ciphertext, new_cipher=True)


_tuple = namedtuple('AESModeNameTuple', 'ECB,CBC,CFB,OFB,CTR,OPENPGP,EAX,CCM,SIV,GCM,OCB')


MODES = _tuple(
    ECB=1, CBC=2, CFB=3, OFB=5, CTR=6,
    OPENPGP=7, EAX=9, CCM=8, SIV=10,GCM=11, OCB=12
)



from multidict import CIMultiDict
from asyncio import AbstractEventLoop
from typing import Any, Union, List, Dict, Optional

from yarl import URL
from aiohttp import ClientSession, ClientResponse, ClientTimeout


__all__ = ['Session', 'EventLoop', 'Response', 'CIMDict', 'URL', 'Timeout', 'URLS']

URL = Optional[Union[str, URL]]
URLS = Union[
    List[URL],
    List[List[Union[Any, Dict[str, Any]]]]
]
EventLoop = Optional[AbstractEventLoop]
Session = Optional[ClientSession]
Response = Optional[ClientResponse]
CIMDict = Optional[CIMultiDict]
Timeout = Union[int, float, ClientTimeout, None]



import os
from concurrent.futures import ThreadPoolExecutor
from typing import (
    Any, Dict, List, Tuple, Union, Optional, Callable, MutableMapping,
)
from urllib.parse import urlsplit, urljoin

from pyxk.utils import (
    chardet,
    make_open,
    LazyLoader,
    get_user_agent,
    download_progress,
    str_conversion_digit,
    read_file_by_generator,
    units_conversion_from_byte
)

from requests.auth import HTTPBasicAuth
from requests.cookies import RequestsCookieJar
from requests.models import Response
from requests.sessions import Session as _Session
from requests.structures import CaseInsensitiveDict
from requests.exceptions import Timeout, InvalidURL, RetryError, ConnectionError as _ConnectionError

_box = LazyLoader("_box", globals(), "rich.box")
_live = LazyLoader("_live", globals(), "rich.live")
_panel = LazyLoader("_panel", globals(), "rich.panel")
_table = LazyLoader("_table", globals(), "rich.table")
_console = LazyLoader("_console", globals(), "rich.console")

Url = Union[str, bytes]
Method = str
Params = Union[Dict[str, str], MutableMapping[str, str], None]
Auth = Union[HTTPBasicAuth, Tuple[Union[str, str]], None]
Cert = Optional[Tuple[Union[str, str]]]
Json = Union[Dict[str, Any], MutableMapping[str, Any], None]
Hooks = Optional[Dict[str, List[Callable]]]
Files = Union[Dict[str, Any], MutableMapping[str, Any], None]
Headers = Union[Dict[str, str], CaseInsensitiveDict, None]
Cookies = Union[dict, RequestsCookieJar, None]
Proxies = Optional[Dict[str, str]]
Data = Union[str, bytes, MutableMapping[str, Any], None]


class Session(_Session):
    """request.Session 重构"""

    def __init__(
        self,
        *,
        headers: Headers = None,
        base_url: Optional[Url] = None,
        user_agent: Optional[str] = None,
    ):
        """request.Session

        :param headers: Headers
        :param base_url: Base Url
        :param user_agent: User Agent
        """
        super().__init__()

        # headers
        headers = CaseInsensitiveDict(headers)
        if not headers.__contains__('User-Agent'):
            headers['User-Agent'] = get_user_agent()

        # user agent
        if user_agent and isinstance(user_agent, str):
            headers['User-Agent'] = user_agent

        # update headers
        self.headers.update(headers)

        # base url
        self._base_url = self._set_base_url(base_url)

        # console
        self._console = _console.Console()

    def request(
        self,
        method: Method,
        url: Url,
        *,
        params: Params = None,
        data: Data = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request

        :param method: 请求方法 'GET', 'POST', 'HEAD', 'OPTIONS', 'DELETE', 'PUT', 'PATCH'
        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        # url 和 method
        url = self.build_url(url)
        if not isinstance(method, str):
            raise TypeError(f'method must be a str, got {type(method).__name__!r}')

        method = method.upper()

        # 最大重试次数 默认10
        max_retries = max_retries if isinstance(max_retries, int) and max_retries > 0 else 1

        # 状态码重试列表
        status_force_list = status_force_list if isinstance(status_force_list, (list, tuple)) else []

        # 指定需要重试的请求方法 默认为所有方法
        method_write_list = [item.upper() for item in method_write_list] \
            if isinstance(method_write_list, (list, tuple)) \
            else ['GET', 'POST', 'HEAD', 'OPTIONS', 'DELETE', 'PUT', 'PATCH']

        # request kwargs
        req_kwargs = dict(
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )

        # 开启request
        result = {'result': False, 'response': None, 'exc': None}

        for index in range(1, max_retries+1):
            result = self._send(
                method=method,
                url=url,
                retries=index,
                show_status=show_status,
                status_force_list=status_force_list,
                method_write_list=method_write_list,
                **req_kwargs
            )
            # 请求成功
            if result['result'] is True:
                break
        else:
            # 请求失败
            if result['result'] is False:
                if isinstance(result['exc'], Exception):
                    raise result['exc']
                else:
                    raise RetryError(
                        f'<Response[{result["response"].status_code}] {url}>, '
                        f'max_retries: {max_retries}'
                    )
        return result['response']

    def _send(
        self,
        method: Method,
        url: Url,
        retries: int,
        show_status: bool,
        status_force_list: List[int],
        method_write_list: List[str],
        **kwargs
    ):
        """send request

        :param method: 请求方法 'GET', 'POST', 'HEAD', 'OPTIONS', 'DELETE', 'PUT', 'PATCH'
        :param url: url
        :param retries: 当前请求重试次数
        :param show_status: 请求显示 rich.status
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :param kwargs: kwargs
        :return: dict
        """
        # rich.status
        rich_status = None
        if show_status:
            retries = f'[{retries}]' if retries > 1 else ''
            rich_status = self.console.status(
                f'{retries}Request <[magenta]{method}[/] [bright_blue u]{url}[/]>',
                spinner='arc'
            )
            rich_status.start()

        result = {'result': False, 'response': None, 'exc': None}

        try:
            # 忽略 requests verify=False 警告
            if not kwargs.get('verify'):
                import urllib3
                urllib3.disable_warnings()

            # request
            response = super().request(method=method, url=url, **kwargs)
            result['response'] = response

            # 请求状态码重试
            if (
                method in method_write_list
                and response.status_code in status_force_list
            ):
                return result
            result['result'] = True
            return result

        # exception
        except (Timeout, _ConnectionError) as exc:
            if method not in method_write_list:
                raise exc
            result['exc'] = exc
            return result

        # 关闭 rich.status
        finally:
            if rich_status:
                rich_status.stop()

    def get(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - get

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='GET',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def post(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - post

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='POST',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def head(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = False,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - head

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='HEAD',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def put(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - put

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='PUT',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def patch(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - patch

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='PATCH',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def delete(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - delete

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='DELETE',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def options(
        self,
        url: Url,
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = None,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
    ) -> Response:
        """request - options

        :param url: url
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :return: requests.Response
        """
        return self.request(
            method='OPTIONS',
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            show_status=show_status,
            max_retries=max_retries,
            status_force_list=status_force_list,
            method_write_list=method_write_list
        )

    def downloader(
        self,
        url: Url,
        method: Method = 'GET',
        *,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        headers: Headers = None,
        cookies: Cookies = None,
        files: Files = None,
        auth: Auth = None,
        timeout: Optional[Union[int, float]] = 5,
        allow_redirects: bool = True,
        proxies: Proxies = None,
        hooks: Hooks = None,
        stream: Optional[bool] = True,
        verify: Optional[Union[bool, str]] = None,
        cert: Cert = None,
        json: Json = None,
        show_status: bool = True,
        max_retries: Optional[int] = 10,
        status_force_list: Optional[List[int]] = None,
        method_write_list: Optional[List[str]] = None,
        output: Optional[str] = None,
        threads: Optional[int] = None,
        restore: bool = False,
        transient: bool = False
    ):
        """request - 下载器

        :param url: url
        :param method: 请求方法 'GET', 'POST', 'HEAD', 'OPTIONS', 'DELETE', 'PUT', 'PATCH'
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param files: files
        :param auth: auth
        :param timeout: 超时时间
        :param allow_redirects: 请求重定向
        :param proxies: 请求代理
        :param hooks: hooks - {'response': [callable, ...]}
        :param stream: 流式相应 大文件请求
        :param verify: verify ssl
        :param cert: 请求证书
        :param json: json
        :param show_status: 请求显示 rich.status
        :param max_retries: 最大重试次数
        :param status_force_list: 状态码重试列表
        :param method_write_list: 请求方式白名单
        :param output: 文件输出路径
        :param threads: 下载线程数量
        :param restore: 文件续传
        :param transient: 转瞬即逝
        :return: requests.Response
        """
        req_kwargs = dict(
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            method_write_list=method_write_list,
            status_force_list=status_force_list,
            max_retries=max_retries,
        )

        # head
        head_response = self.head(url=url, show_status=show_status, **req_kwargs)

        req_kwargs['url'] = url
        req_kwargs['method'] = method
        req_kwargs['show_status'] = False

        # content length
        content_length = str_conversion_digit(head_response.headers.get('Content-Length')).result

        # invalid content length
        if not content_length:
            threads, content_length = 1, None

        # output
        output = os.path.normpath(os.path.abspath(output)).replace('\\', '/') \
            if output and isinstance(output, str) else None

        # threads
        threads = 1 \
            if not output or not isinstance(threads, int) or threads <= 0 \
            else threads

        # not output - exit
        if not output:
            panel = self._repr_downloader(
                url=url,
                output=output,
                length=content_length,
                threads=threads,
                file_type=head_response.headers.get('Content-Type'),
                response=head_response
            )

            self.console.print(panel['panel'])
            return head_response

        # 多线程下载
        panel = self._repr_downloader(
            url=url,
            output=output,
            length=content_length,
            threads=threads,
            file_type=head_response.headers.get('Content-Type'),
            response=head_response,
            add_progress=True
        )

        # output_list, threads_pool
        output_list, pool = [], ThreadPoolExecutor(max_workers=8)

        with _live.Live(panel['panel'], console=self.console, transient=transient):
            for index in range(threads):
                # start_byte, end_byte
                start_byte, end_byte = None, None

                if content_length:
                    start_byte = index * (content_length // threads)
                    end_byte = start_byte + content_length // threads - 1

                    if index == threads - 1:
                        end_byte = content_length - 1

                # threads output
                thread_output = f'{output}.{index + 1}_{threads}.tmp' if threads > 1 else output
                output_list.append(thread_output)

                # 开启线程下载
                pool.submit(
                    self._downloader_chunk,
                    output=thread_output,
                    progress=panel['progress'],
                    restore=restore,
                    start_byte=start_byte,
                    end_byte=end_byte,
                    **req_kwargs
                )
            pool.shutdown()
            # 合并文件
            self._file_merge(output, output_list)

        return head_response

    def _downloader_chunk(
        self,
        output: str,
        progress,
        restore: bool = False,
        start_byte: Optional[int] = None,
        end_byte: Optional[int] = None,
        **kwargs
    ):
        """下载器 - 下载块

        :param output: 文件下载路径
        :param progress: rich.progress.Progress
        :param restore: 文件续传
        :param start_byte: 文件下载开始部分
        :param end_byte: 文件下载结束部分
        :param kwargs: request params
        """
        # total
        total = end_byte - start_byte \
            if isinstance(start_byte, int) and isinstance(end_byte, int) \
            else None

        # restore
        restore = restore if isinstance(total, int) else False

        # 文件续传
        completed, chunk_size = 0, 1024

        if restore and os.path.isfile(output):
            completed = os.path.getsize(output)
            if completed > total + 1:
                restore, completed = False, 0
            else:
                start_byte += completed
        else:
            restore = False

        # rich.progress.Progress 创建 task
        task = progress.add_task(description='', total=total, completed=completed)

        # request headers
        if isinstance(start_byte, int) and isinstance(end_byte, int):
            headers = CaseInsensitiveDict(kwargs.pop('headers', {}))
            headers.update({'Range': f'bytes={start_byte}-{end_byte}'})
            kwargs['headers'] = headers

        # 开启流式下载
        response = self.request(**kwargs)
        # 状态码 416 请求超出 range范围
        if response.status_code == 416:
            return

        with make_open(output, "ab" if restore else "wb") as write_file_obj:
            for chunk in response.iter_content(chunk_size=chunk_size):
                write_file_obj.write(chunk)
                progress.update(task, advance=chunk_size)

    @staticmethod
    def _file_merge(output: str, merge_list: list):
        """多线程下载后文件合并

        :param output: 文件合并后的路径
        :param merge_list: 需要合并的文件列表
        """
        # 判断文件是否存在
        for file in merge_list:
            if not os.path.isfile(file):
                return
        os.rename(merge_list.pop(0), output)

        # 拼接文件
        if not merge_list:
            return

        with make_open(output, 'ab') as write_file_obj:
            for file in merge_list:
                for _chunk in read_file_by_generator(file, 'rb'):
                    write_file_obj.write(_chunk)

                # 合并完成 删除文件
                os.remove(file)

    @staticmethod
    def _repr_downloader(
        url: Url,
        output: Optional[str],
        length: Optional[int],
        threads: int,
        file_type: Optional[str],
        response: Response,
        add_progress: bool = False,
    ):
        """显示downloader结果 - rich.panel.Panel"""
        result = {'panel': None, 'progress': None}

        table = _table.Table(show_header=False, box=_box.SIMPLE_HEAD)
        table.add_column(justify='left', overflow='fold')
        table.add_row(f'<[cyan]Response[/] [{response.status_code}]> [blue u]{url}[/]')
        table.add_section()

        # invalid content length
        if not length:
            table.add_row('[red]content-length is not available![/]')
            table.add_section()

        table.add_row(f'[green]filetype[/]: [yellow]{file_type}[/]')
        table.add_row(f'[green]filesize[/]: [yellow]{units_conversion_from_byte(length).result}[/] ({length})')
        table.add_row(f'[green]filename[/]: [blue]{output}[/]')

        # 添加rich.progress.Progress
        if add_progress:
            table.add_section()
            progress = download_progress()
            table.add_row(progress)
            result['progress'] = progress

        panel = _panel.Panel(
            table,
            title='[red b]Downloader[/]',
            title_align='center',
            border_style='bright_blue',
            expand=False,
            subtitle=f'[dim i]thread: {threads}',
            subtitle_align='right'
        )

        result['panel'] = panel
        return result

    def build_url(self, url: Url) -> Url:
        """构造完整url"""
        # 无效 url
        if not url or not isinstance(url, (str, bytes)):
            raise InvalidURL(f'url must be a str or bytes, got {url!r}')

        url = self._get_string_url(url)

        # 拼接完整 url
        if self.base_url and not self.is_absolute_url(url):
            return urljoin(self.base_url, url)

        return url

    def is_absolute_url(self, url: Url) -> bool:
        """判断 url 是否为绝对路径

        :param url: url
        :return: bool
        """
        if not url or not isinstance(url, (str, bytes)):
            return False

        url = urlsplit(self._get_string_url(url))

        return bool(url.scheme and url.netloc)

    def _set_base_url(self, url: Url) -> Optional[str]:
        """设置 base url

        :param url: url
        :return: Optional[str]
        """
        url = self._get_string_url(url)
        if self.is_absolute_url(url):
            return url
        return None

    @staticmethod
    def _get_string_url(url: Url) -> str:
        """获取字符串url"""
        if not isinstance(url, bytes):
            return url
        return url.decode(chardet(url).encoding)

    @property
    def base_url(self):
        """Base Url"""
        return self._base_url

    @base_url.setter
    def base_url(self, value):
        self._base_url = self._set_base_url(value)

    @property
    def console(self):
        """rich.console.Console"""
        return self._console

    @property
    def user_agent(self):
        """User-Agent"""
        return self.headers.get('User-Agent')

    @user_agent.setter
    def user_agent(self, value):
        self.headers.update({'User-Agent': value})



import click
from pyxk import get_user_agent
from pyxk.m3u8 import load_url, load_content


@click.group(invoke_without_command=False, chain=False)
@click.option("-o", "--output", "output", type=str, default=None, help="M3U8存储路径")
@click.option("--reload", is_flag=True, help="重载m3u8资源")
@click.option("--reserve", is_flag=True, help="保留m3u8资源")
@click.option("-h", "--headers", "headers", type=(str, str), multiple=True, help="Request Headers")
@click.option("--no-verify", "verify", is_flag=True, default=True, help="Request Verify")
@click.option("-l", "--limit", "limit", type=int, default=16, help="下载并发量")
@click.option("-ua", "--user-agent", "user_agent", type=str, default=None, help="User-Agent")
@click.pass_context
def main(ctx, output, reload, reserve, headers, verify, limit, user_agent):
    """m3u8下载器"""
    ctx.obj = {}
    # 将参数传递给子命令
    if ctx.invoked_subcommand:
        ctx.obj.update(ctx.params)


@main.command
@click.pass_obj
@click.argument("content", type=str, metavar="<File>")
@click.option("-u", "--url", "_url", type=str, default=None, help="m3u8 url")
@click.option("-o", "--output", "output", type=str, default=None, help="M3U8存储路径")
@click.option("--reload", is_flag=True, help="重载m3u8资源")
@click.option("--reserve", is_flag=True, help="保留m3u8资源")
@click.option("-h", "--headers", "headers", type=(str, str), multiple=True, help="Request Headers")
@click.option("--no-verify", "verify", is_flag=True, default=True, help="Request Verify")
@click.option("-l", "--limit", "limit", type=int, default=None, help="下载并发量")
@click.option("-ua", "--user-agent", "user_agent", type=str, default=None, help="User-Agent")
def file(obj, content, _url, output, reload, reserve, headers, verify, limit, user_agent):
    """使用m3u8文件下载资源"""
    load_content(
        content=content,
        url=_url,
        output=output or obj["output"],
        reload=reload,
        reserve=reserve,
        headers=dict(headers or obj["headers"]),
        verify=verify,
        limit=limit or obj["limit"],
        user_agent=get_user_agent(user_agent or obj["user_agent"])
    )


@main.command
@click.pass_obj
@click.argument("_url", type=str, metavar="<Url>")
@click.option("-o", "--output", "output", type=str, default=None, help="M3U8存储路径")
@click.option("--reload", is_flag=True, help="重载m3u8资源")
@click.option("--reserve", is_flag=True, help="保留m3u8资源")
@click.option("-h", "--headers", "headers", type=(str, str), multiple=True, help="Request Headers")
@click.option("--no-verify", "verify", is_flag=True, default=True, help="Request Verify")
@click.option("-l", "--limit", "limit", type=int, default=None, help="下载并发量")
@click.option("-ua", "--user-agent", "user_agent", type=str, default=None, help="User-Agent")
def url(obj, _url, output, reload, reserve, headers, verify, limit, user_agent):
    """使用m3u8链接下载资源"""
    load_url(
        url=_url,
        output=output or obj["output"],
        reload=reload,
        reserve=reserve,
        headers=dict(headers or obj["headers"]),
        verify=verify,
        limit=limit or obj["limit"],
        user_agent=get_user_agent(user_agent or obj["user_agent"])
    )



from pyxk.requests.sessions import Session
from pyxk.requests.api import delete, get, head, options, patch, post, put, request, downloader


__all__ = ["Session", "delete", "get", "head", "options", "patch", "post", "put", "request", "downloader"]
